/* $Workfile: OsaLinuxSpi.c $   $Revision: 11 $    $Modtime: 10/11/08 1:40p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007-2008, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places


// Only include in DaVinci
#ifdef DAVINCI

#include <linux/syscalls.h>
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/smp_lock.h>


#include <asm/uaccess.h>
#include <asm/system.h>

#include "OsaLinuxSpi.h"

#define OD_CONFIG_SPI_DEBUG 1
#if defined(OD_CONFIG_SPI_DEBUG)
#define DEBUG_SPI(fmt, arg...)  printk(KERN_EMERG fmt , ##arg);
#else
#define DEBUG_SPI(fmt, arg...)
#endif
 
#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
 
#include <linux/kernel.h>   /* printk() */
#include <linux/slab.h>     /* kmalloc() */
#include <linux/fs.h>       /* everything... */
#include <linux/errno.h>    /* error codes */
#include <linux/types.h>    /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>    /* O_ACCMODE */
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <linux/delay.h>
//#include <linux/time.h>
#include <linux/semaphore.h>
#include <asm/uaccess.h>    /* copy_*_user */
//
//#include <asm/system.h>     /* cli(), *_flags */
//#include <asm/io.h>
#include <asm/semaphore.h>
//#include <asm/arch/hardware.h>
//#include <asm/hardware/clock.h>
 
#include <linux/ioctl.h>
#include <linux/time.h>
  
 
// Writes to SPI device channel (aka CS, chip select) from supplied kernel buffer // count bytes. Returns number of words really written.
ssize_t spi_write_internal(unsigned int channel, unsigned short * buf, size_t count);
 
 
// debug helper - dumps registers' content into printk void spi_dump_registers(void);  
/*******************************************************************************************************
 
    FOR DETAILS OF OPERATIONS SEE TI SPRUE32 (SPI data sheet)
   
*******************************************************************************************************/
//#define DEBUG_SPI_V2 1
#ifdef  DEBUG_SPI_V2
#define DBGENTER printk("DEBUG_SPI_V2: Entered %s\n", __FUNCTION__)
#define DBGEXIT  printk("DEBUG_SPI_V2: Exited %s\n", __FUNCTION__)
#else
#define DBGENTER
#define DBGEXIT
#endif
 
//#define SPI_DBG(x...) printk(x)
#define SPI_DBG(x...)
 
//
// REGISTER ADDRESS
//
#define SPIGCR0     __REG(0x1C66800)
#define SPIGCR1     __REG(0x1C66804)
#define SPIINT      __REG(0x1C66808)        // RXINTEN, BIT 8 1 is enabled
#define SPILVL      __REG(0x1C6680C)        // RXINTLVL, BIT 8 0 if mapped to int0, 1 if mapped to 1
#define SPIFLG      __REG(0x1C66810)        // RXINTFLAG, BIT 8 1 if rx int occurred
#define SPIPC0      __REG(0x1C66814)       
#define SPIPC1      __REG(0x1C66818)        // OK-S I suspect this register exists, and that
                                            // it controls pin direction.Probably same format
                                            // as SPIPC0.  Guessing 1 is out 0 is in.  
#define SPIPC2      __REG(0x1C6681C)
#define SPIPC3      __REG(0x1C66820)
#define SPIPC4      __REG(0x1C66824)
#define SPIPC5      __REG(0x1C66828)
#define SPIPC6      __REG(0x1C6682C)        // OK-S probably exists.  TI code indicates it may be
                                            // open drain control.
#define SPIPC7      __REG(0x1C66830)        // OK-S probably exists.  TI code indicate it may be
                                            // pull up control
#define SPIDAT1     __REG(0x1C6683C)
#define SPIBUF      __REG(0x1C66840)
#define SPIEMU      __REG(0x1C66844)
#define SPIDELAY    __REG(0x1C66848)
#define SPIDEF      __REG(0x1C6684C)
#define SPIFMT0     __REG(0x1C66850)
#define SPIFMT1     __REG(0x1C66854)
#define SPIFMT2     __REG(0x1C66858)
#define SPIFMT3     __REG(0x1C6685C)
#define SPIIVEC0    __REG(0x1C66860)
#define SPIIVEC1    __REG(0x1C66864)
 
//#define IOCTL_SPI_0_DATAFORMAT  17
//#define IOCTL_SPI_1_DATAFORMAT  18
 
#define PINMUX0         __REG(0x01c40000)
#define HARD_DRIVE_DIR  16
#define PINMUX1         __REG(0x01c40004)
#define SPI_PIN_MUX 8
#define MDSTAT  IO_ADDRESS(0x01C41800)
#define MDCTL   IO_ADDRESS(0x01C41A00)
#define PTCMD       __REG(0x01C41120)
#define PDSTAT      __REG(0x01C41200)
#define PDCTL1      __REG(0x01C41304)
#define PTSTAT      __REG(0x01C41128)
#define EPCPR       __REG(0x01C41070)
 
// general
#define BIT(x) x
#define SET(reg, bit)    reg |= (1 << bit)
#define CLEAR(reg, bit)  reg &= ~(1 << bit)
#define ISSET(reg, bit)  (reg & (1 << bit))
 
// SPIGCR0 declarations
#define RESET           BIT(0)
 
// SPIGCR1 declarations
#define SPIENA          BIT(24)
#define LOOPBACK        BIT(16)
#define CLKMOD          BIT(1)
#define MASTER          BIT(0)
 
// SPIFLG declarations
#define RXINTFLAG       BIT(8)
#define OVRNINTFLG      BIT(6)
#define BITERRFLG       BIT(4)
 
// SPIPC0 declarations
#define DIFUN           BIT(11)
#define DOFUN           BIT(10)
#define CLKFUN          BIT(9)
#define EN1FUN          BIT(1)
#define EN0FUN          BIT(0)
 
// SPIDAT1 declarations
#define CSHOLD          BIT(28)
#define DFSEL           BIT(24)
#define CSNR            BIT(16)
 
// SPIBUF declarations
#define RXEMPTY         BIT(31)
#define RXOVR           BIT(30)
#define TXFULL          BIT(29)
#define BITERR          BIT(28)
#define LCSNR           BIT(16)
 
// SPIDELAY declarations
#define C2TDELAY        BIT(24)
#define T2CDELAY        BIT(16)
 
//SPIDEF declarations
#define EN1DEF          BIT(1)
#define EN0DEF          BIT(0)
 
// SPIFMTn declarations
#define SHIFTDIR        BIT(20)
#define POLARITY        BIT(17)
#define PHASE           BIT(16)
#define PRESCALE        BIT(8)
#define CHARLEN         BIT(0)
 
static struct timeval t0;
static struct timeval t1;
 
#define  T_START()   do_gettimeofday(&t0)
#define  T_END()     do_gettimeofday(&t1)

/* 
static unsigned int timeDiff(void)
{
    return ((t1.tv_sec - t0.tv_sec)*1000000 + t1.tv_usec - t0.tv_usec);
}
*/ 
  
//
struct SPI_CHANNEL_DATA
{
    unsigned long   inter_char_interval_mks;
};
 
//
#define SPI_CHANNEL_COUNT   2
 
//
struct SPI_GLOBAL_DATA
{
    struct SPI_CHANNEL_DATA devices[SPI_CHANNEL_COUNT];
    struct semaphore        spiAccess;
} spi_global_data;
 
 
int bSpiInitialized = 0;


static spinlock_t ms_accessInterlocked;

 
//
void spi_dump_registers(void)
{
    SPI_DBG("SPIDAT1 = %x, SPIBUF =%x, SPIINT =%x, SPIFLG =%x\n",
             (unsigned int)SPIDAT1,
             (unsigned int)SPIBUF,
             (unsigned int)SPIINT,
             (unsigned int)SPIFLG);
 
    SPI_DBG("SPIFMT0 = %x, SPIFMT1 = %x, SPIDEF = %x\n",
             (unsigned int)SPIFMT0,
             (unsigned int)SPIFMT1,
             (unsigned int)SPIDEF);
 
    SPI_DBG("SPIPC0 = %x, SPIPC2 = %x\n",
             (unsigned int)SPIPC0,
             (unsigned int)SPIPC2);
 
    SPI_DBG( "\nSPIGCR1 (%x) SPIPC0(%x) SPIGCR0 (%x) SPIPC2 (%x)\n",
            (unsigned int)SPIGCR1,
            (unsigned int)SPIPC0,
            (unsigned int)SPIGCR0,
            (unsigned int)SPIPC2 );
 
    SPI_DBG( "SPIPC0(%x) SPIPC1(%x) SPIPC2(%x) SPIPC3(%x) SPIPC4(%x) SPIPC5(%x) SPIPC6(%x) SPIPC7(%x)\n",
            (unsigned int)SPIPC0,
            (unsigned int)SPIPC0,
            (unsigned int)SPIPC2,
            (unsigned int)SPIPC3,
            (unsigned int)SPIPC4,
            (unsigned int)SPIPC5,
            (unsigned int)SPIPC6,
            (unsigned int)SPIPC7 );
 
    SPI_DBG( "PINMUX0 (%x) PINMUX1(%x)\n", (unsigned int)PINMUX0, (unsigned int)PINMUX1 );

}
 
// init SPI
static int __init spi_controller_init(void) {
 

    DBGENTER;
    // reset SPI
    CLEAR(SPIGCR0, RESET);
    // leave reset state
    SET(SPIGCR0, RESET);
 
    // enable SPI
    SPIGCR1 = (1 << CLKMOD) | (1 << MASTER);
 
    // enable SPI pins
    SPIPC0 = (1 << DIFUN) | (1 << DOFUN) | (1 << CLKFUN) | (1 << EN1FUN) | (1 << EN0FUN);
    //  SPIPC1 = DO_CLK_EN0_EN1;
    //  SPIPC2 = ENABLE_PINS;
    //  SPIPC3 = ENABLE_PINS;
    //  SPIPC6 = ENABLE_PINS;
    //  SPIPC7 = ENABLE_PINS;
    //  SPIGCR1 = ENABLE_CKL_MSTR;
 
    // specify default cs pattern (both become inactive of no active transaction is present)
    SPIDEF = (1 << EN1DEF) | (1 << EN0DEF);
 
    // disable interrupts from SPI
    SPIINT = 0;

    // pre configure formats for both channels
    // all formats can be changed throuh IOCTL call

    // 0 device is used for FALC communication
    // msb first, low polarity, ...
    //SPIFMT0 = (0 << SHIFTDIR) | (0 << POLARITY) | (1 << PHASE) | (0x10 << PRESCALE) | (16 << CHARLEN); // Faster
    SPIFMT0 = (0 << SHIFTDIR) | (0 << POLARITY) | (1 << PHASE) | (0x18 << PRESCALE) | (16 << CHARLEN); // 

    // 1 device is not used (yet)
    SPIFMT1 = (0 << SHIFTDIR) | (0 << POLARITY) | (1 << PHASE) | (0x30 << PRESCALE) | (8 << CHARLEN);
 
    // reset SPI
    //CLEAR(SPIGCR0, RESET);
    // leave reset state
    //SET(SPIGCR0, RESET);
   
    // enable SPI
    SET(SPIGCR1, SPIENA);

    return 0;
}

//
static unsigned int setup_transaction(int channel)
{
  unsigned long spidat1 = 0;
  
  DBGENTER;
  
  // pre cond
  if(channel != 0 && channel != 1) {
    printk(KERN_ALERT "%s (%d): LOGIC ERROR", __func__, __LINE__ );
    // TODO: generate oops!
  }
  
  // disable before configuration
  CLEAR(SPIGCR1, SPIENA);
  
  // select format (channel X ~ format X)
  spidat1 |= (channel << DFSEL);
  
  // set CS
  if(channel)
    spidat1 |= (1 << CSNR);
  else
    spidat1 |= (2 << CSNR);
  
  // CS delays
  SPIDELAY = 0;
  
  // SPIDELAY does not seem to matter since we set CSHOLD
  //SPIDELAY |= (0xF << C2TDELAY);
  //SPIDELAY |= (0xF << T2CDELAY);
  SPIDELAY |= (0x2 << C2TDELAY);
  SPIDELAY |= (0x2 << T2CDELAY);
  
  // CS HOLD setup
  spidat1 |= (1 << CSHOLD);
  
  // enable after configuration
  SET(SPIGCR1, SPIENA);
  
  return spidat1;
}
 
static unsigned int end_transaction(int channel)
{
  DBGENTER;
  // disable before configuration
  CLEAR(SPIGCR1, SPIENA);
  
  // clear CS & CSHOLD
  SPIDAT1 = (3 << CSNR) | (0 << CSHOLD);
  
  // enable after configuration
  SET(SPIGCR1, SPIENA);
  return 0;
}
 
 
//
// Single Word exchange
//
static unsigned short spi_raw_rw_word( int channel, unsigned long spidat1, unsigned int * done, unsigned long inter_char_interval_mks )
{
  unsigned long status = 0;
  unsigned short result = 0;
  unsigned long nBusyCount = 0;
  
  DBGENTER;
  *done = 0;
  
  //printk("SPIDAT1 = %x\n", (unsigned int)spidat1);
  T_START();
  //SPI_DBG("Commiting %x to SPIDAT1\n", (unsigned int)spidat1 );
  SPIDAT1 = spidat1;
  
  do {
    if(!ISSET(SPIFLG, RXINTFLAG)) {
      nBusyCount++;
      continue;
    } else {
      *done = 1;
    }
    
    status = SPIBUF;
  } while(*done != 1);
  
  T_END();
  //printk("Exchange time diff = %d mks\n", timeDiff());
  
  udelay(inter_char_interval_mks); //50 for PC!!!!
  
  if(*done == 0) {
    printk(KERN_WARNING "%s\n", "Not done!!!!");
    spi_dump_registers();
  }
  
  result = (unsigned short)(status & 0xFFFF);
  SPI_DBG("Read back 0x%.4X  (nBusyCount=%d)\n", result, nBusyCount);
  
  return result;
}
 
 
// write to SPI device channel
ssize_t spi_write_internal(unsigned int channel, unsigned short * buf, size_t count)
{
  unsigned int i = 0;
  unsigned short spiWord = 0;
  unsigned long spidat1 = 0;
  unsigned int done = 0;
  ssize_t retval = -1; // error
  
  DBGENTER;
  SPI_DBG( "%s (%d): Write %d words to SPI\n", __func__, __LINE__, count );
  
  if (down_interruptible(&spi_global_data.spiAccess))
    return -ERESTARTSYS;
  
  spidat1 = setup_transaction(channel);
  
  for (i = 0; i < count; i++ ) {
    spiWord = buf[i];
    
    spidat1 &= 0xFFFF0000; // clear data part of register
    spidat1 |=spiWord;
    
    buf[i] = (unsigned short) spi_raw_rw_word(channel, spidat1, &done, spi_global_data.devices[channel].inter_char_interval_mks);
    if(done != 1)
    {
      printk( "%s (%d): Error writing word from channel %d\n", __func__, __LINE__, channel);
      break; // error
    }
  }
  
  retval = i;
  end_transaction(channel);
  up( &spi_global_data.spiAccess );
  
  return retval;
}

 
static void __init spi_board_setup_psc(unsigned int domain, unsigned int id, char enable)
{
  volatile unsigned int *mdstat = (unsigned int *)((int)MDSTAT + 4 * id);
  volatile unsigned int *mdctl = (unsigned int *)((int)MDCTL + 4 * id);
  
  DBGENTER;       
  if (enable) {
    *mdctl |= 0x00000003;   /* Enable Module */
  } else {
    *mdctl &= 0xFFFFFFF2;   /* Disable Module */
  }
  
  
  if ((PDSTAT & 0x00000001) == 0) {
    PDCTL1 |= 0x1;
    PTCMD = (1 << domain);
    while ((((EPCPR >> domain) & 1) == 0)) ;
    
    
    PDCTL1 |= 0x100;
    while (!(((PTSTAT >> domain) & 1) == 0)) ;
  } else {
    PTCMD = (1 << domain);
    while (!(((PTSTAT >> domain) & 1) == 0)) ;
  }
  
  if (enable) {
    while (!((*mdstat & 0x0000001F) == 0x3)) ;
  } else {
    while (!((*mdstat & 0x0000001F) == 0x2)) ;
  }
}

static void spi_cleanup_module(void)
{
}

/******************************************************************************
  Open the SPI device
    return 0=>closed, 1=>open
******************************************************************************/
OsaBool32 OsaLinuxSpiIsOpen(void)
{
  //DEBUG_SPI("OsaLinuxSpiOpen: Is SPI open? (bSpiInitialized=%d)\n", bSpiInitialized);
  return (bSpiInitialized) ? OSA_TRUE : OSA_FALSE;
}


/******************************************************************************
  Open the SPI device
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiOpen(void)
{  
  DBGENTER;
  SPI_DBG("OsaLinuxSpiOpen() - Initializing the SPI module\n");

  ms_accessInterlocked = ms_accessInterlocked;
  spin_lock_init(&ms_accessInterlocked);
  
  /*
  T_START();
  msleep(1);
  T_END();
  printk("Test diff (1 ms)= %d ms\n", timeDiff()/1000);
  T_START();
  msleep(1000);
  T_END();
  printk("Test diff (1s)= %d s\n", timeDiff()/1000000);
  */
  
  // TODO ! !!!! Check code below!
  
  // Make sure we have some power and food for the SPI module
  spi_board_setup_psc(DAVINCI_GPSC_ARMDOMAIN, DAVINCI_LPSC_SPI, 1);
  // Disable the HDDIR pin
  PINMUX0 &= ~(1<<HARD_DRIVE_DIR);
  // Make sure the SPI pins are activate
  PINMUX1 |= 1<<SPI_PIN_MUX;
  
  // init cross-channel mutex
  init_MUTEX( &spi_global_data.spiAccess );
  
  spi_global_data.devices[0].inter_char_interval_mks = 0; // Falc
  spi_global_data.devices[1].inter_char_interval_mks = 0;  // touch panel (not used in Alvis)
    
  // init SPI
  spi_controller_init();
  spi_dump_registers();
  
  bSpiInitialized = 1;
  return 0;
  
}

/******************************************************************************
  Close the SPI device. Not sure why this would be needed, OTX driver should
  be active a all times. 
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiClose(void)
{
  spi_cleanup_module();

  bSpiInitialized = 0;
  return 0;
}


/******************************************************************************
  Read a single byte from the SPI
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiRead8(OsaUint16 nAddress, OsaUint8 *nData)
{
  OsaUint16 buf[2];
  ssize_t rc;
  SPI_DBG("In OsaLinuxSpiRead8()\n");

  buf[0] = nAddress | 0x8000;  // 0x80 since the Falc SPI needs a one as the first bit to indicate a write operation
  buf[1] = 0x0000; // Dummy value
  SPI_DBG("0:%.4X 1:%.4X \n", buf[0], buf[1]);
  rc = spi_write_internal(0, (unsigned short *)buf, 2);

  SPI_DBG("spi_write_internal returned %d (nData=%.4X)\n", rc, buf[1]);
  *nData = (OsaUint8)(buf[1] >> 8);

  return (rc > 0) ? 0 : -1;
}


/******************************************************************************
  Write a single byte to the SPI.
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiWrite8(OsaUint16 nAddress, OsaUint8 nData)
{
  OsaUint16 buf[2];
  ssize_t rc;

  SPI_DBG("In OsaLinuxSpiWrite8() nAddress=0x%.4X, nData=0x%.2X\n", nAddress, nData);

  buf[0] = nAddress & ~0x8000; // ~0x80 since the Falc SPI needs a zero as the first bit to indicate a write operation
  buf[1] = (OsaUint16)nData << 8; 
  SPI_DBG("0:%.4X 1:%.4X\n", buf[0], buf[1]);

  rc = spi_write_internal(0, (unsigned short *)buf, 2);

  SPI_DBG("spi_write_internal returned %d\n", rc);
  return (rc > 0) ? 0 : -1;
}

/******************************************************************************
  @devnote Note that the Osa Interlocked functions are only guaranteed to be
  interlocked in relation to eachother.
******************************************************************************/
OsaUint8 OsaLinuxSpiInterlockedModifyRegBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits)
{
  OsaUint8 nValue;
  __u32 nSavedFlags;

  OSA_ASSERT(!(spin_is_locked(&ms_accessInterlocked)));
  spin_lock_irqsave(&ms_accessInterlocked, nSavedFlags);

  OsaLinuxSpiRead8((OsaUint16)pRegister, &nValue);

  // Clear the bits we modify
  nValue &= ~nBitsToModify;
  nValue |= nValueOfBits;

#if DBG
  // Make sure the value doesn't modify bits it shouldn't
  OSA_ASSERT((nValueOfBits & ~nBitsToModify) == 0);
#endif

  OsaLinuxSpiWrite8((OsaUint16)pRegister, nValue);

  spin_unlock_irqrestore((spinlock_t *) &ms_accessInterlocked, 
                         (unsigned long) nSavedFlags);

  return nValue;
}



#endif // DAVINCI
